from __future__ import annotations

import abc
from collections.abc import Callable, Sequence
from logging import getLogger
from typing import Optional

import numpy as np
from numpy.typing import ArrayLike
from pandas import DataFrame, Series
from scipy.interpolate import interp1d
from scipy.special import wofz

import riip.dataframe

from .formulas import formulas_cython_dict, formulas_numpy_dict

logger = getLogger(__package__)


def _ensure_positive_imag(x: ArrayLike) -> np.ndarray:
    """If the imaginary part of x is negative, change it to zero."""
    _x = np.asarray(x, dtype=np.complex128)
    return _x.real + 1j * _x.imag * (_x.imag > 0)


class AbstractMaterial(metaclass=abc.ABCMeta):
    """Abstract base class for materials"""

    @abc.abstractmethod
    def __init__(self, *args) -> None:
        self.label = ""

    @abc.abstractmethod
    def n(self, wls: ArrayLike) -> np.ndarray:
        """Retrun refractive index at wavelength wls [μm]"""
        return np.asarray(wls, dtype=np.float64)

    @abc.abstractmethod
    def k(self, wls: ArrayLike) -> np.ndarray:
        """Return extinction coefficient at wavelength wls [μm]"""
        return np.asarray(wls, dtype=np.float64)

    @abc.abstractmethod
    def eps(self, wls: ArrayLike) -> np.ndarray:
        """Return permittivity at wavelength wls [μm]"""
        return np.asarray(wls, dtype=np.float64)

    def bound_check(self, wl: ArrayLike, nk: str) -> None:
        pass

    def plot(
        self,
        wls: Sequence | np.ndarray,
        comp: str = "n",
        fmt1: Optional[str] = "-",
        fmt2: Optional[str] = "--",
        **kwargs,
    ) -> None:
        """Plot refractive index, extinction coefficient or permittivity.

        Args:
            wls (Sequence | np.ndarray): Wavelength coordinates to be plotted [μm].
            comp (str): 'n', 'k' or 'eps'
            fmt1 (Optional[str]): Plot format for n and Re(eps).
            fmt2 (Optional[str]): Plot format for k and Im(eps).
        """
        import matplotlib.pyplot as plt

        kwargs.setdefault("lw", 4)
        kwargs.setdefault("ms", 8)
        if comp == "n":
            ns = self.n(wls)
            plt.plot(wls, ns, fmt1, label=self.label, **kwargs)
            plt.ylabel(r"$n$")
        elif comp == "k":
            ks = self.k(wls)
            plt.plot(wls, ks, fmt2, label=self.label, **kwargs)
            plt.ylabel(r"$k$")
        elif comp == "eps":
            eps = self.eps(wls)
            (line,) = plt.plot(wls, eps.real, fmt1, label=self.label, **kwargs)
            color = line.get_color()
            plt.plot(wls, eps.imag, fmt2, color=color, **kwargs)
            plt.ylabel(r"$\varepsilon$")
        plt.xlabel(r"$\lambda$ $[\mathrm{\mu m}]$")
        plt.legend()


class RiiMaterial(AbstractMaterial):
    """This class provide dispersion formula defined in refractiveindex.info database.

    Attributes:
        catalog: Catalog of data.
        raw_data: The experimental data set.
    """

    def __init__(
        self, id: int, catalog: DataFrame, raw_data: DataFrame, bound_check: bool = True
    ) -> None:
        """Initialize RiiMaterial

        Args:
            id (int): ID number
            catalog (DataFrame): catalog of Rii_Pandas DataFrame.
            raw_data (DataFrame): raw_data of Rii_Pandas DataFrame.
            bound_check (bool): True if bound check should be done. Defaults to True.
        """
        self.catalog: Series = catalog.loc[id]
        # raw_data becomes a Series if it has only 1 row.
        self.raw_data: Series | DataFrame = raw_data.loc[id]
        self.f = int(self.catalog["formula"])
        if self.f > 0:
            self.cs = self.raw_data["c"].to_numpy()[:24]
            self.formula = lambda x: formulas_numpy_dict[self.f](x, self.cs)
        self.label = f"{self.catalog['book']} {self.catalog['page']}"
        self.bound_check_flag = bound_check
        self.__n = self._func_n()
        self.__k = self._func_k()

    def bound_check(self, wl: ArrayLike, nk: str) -> None:
        """Raise ValueError if wl is out of bounds"""
        _x = np.atleast_1d(wl)
        if not self.bound_check_flag:
            return
        if nk == "n":
            wl_min = self.catalog["wl_n_min"]
            wl_max = self.catalog["wl_n_max"]
        elif nk == "k":
            wl_min = self.catalog["wl_k_min"]
            wl_max = self.catalog["wl_k_max"]
        elif nk == "nk":
            wl_min = self.catalog["wl_min"]
            wl_max = self.catalog["wl_max"]
        else:
            raise ValueError("nk must be 'n', 'k', or 'nk'.")
        x_min = min(_x)
        x_max = max(_x)
        if x_min < wl_min or x_max > wl_max:
            raise ValueError(
                f"Wavelength [{x_min} {x_max}] is out of bounds [{wl_min} {wl_max}][um]"
            )

    def _func_n(self) -> Callable:
        tabulated = self.catalog["tabulated"]
        if self.f > 0:
            if self.f <= 20:
                return self.formula
            else:
                return lambda x: np.sqrt(_ensure_positive_imag(self.formula(x))).real
        elif "n" in tabulated:
            num_n = self.catalog["num_n"]
            if num_n == 1:
                return lambda x: self.raw_data["n"] * np.ones_like(x)
            elif num_n < 4:
                val = np.mean(self.raw_data["n"])
                return lambda x: val * np.ones_like(x)
            else:
                return interp1d(
                    self.raw_data["wl_n"].to_numpy()[:num_n],
                    self.raw_data["n"].to_numpy()[:num_n],
                    kind="cubic",
                    bounds_error=False,
                    fill_value="extrapolate",
                    assume_sorted=True,
                )
        else:
            logger.warning("Refractive index is missing and set to zero.")
            return lambda x: np.zeros_like(x)

    def _func_k(self) -> Callable:
        tabulated = self.catalog["tabulated"]
        if "k" in tabulated:
            num_k = self.catalog["num_k"]
            if num_k == 1:
                return lambda x: self.raw_data["k"] * np.ones_like(x)
            elif num_k < 4:
                val = np.mean(self.raw_data["k"])
                return lambda x: val * np.ones_like(x)
            else:
                return interp1d(
                    self.raw_data["wl_k"].to_numpy()[:num_k],
                    self.raw_data["k"].to_numpy()[:num_k],
                    kind="cubic",
                    bounds_error=False,
                    fill_value="extrapolate",
                    assume_sorted=True,
                )
        else:
            formula = int(self.catalog["formula"])
            if formula > 20:
                return lambda x: np.sqrt(_ensure_positive_imag(self.formula(x))).imag
            else:
                logger.warning("Extinction coefficient is missing and set to zero.")
                return lambda x: np.zeros_like(x)

    def n(self, wl: ArrayLike) -> np.ndarray:
        """Return refractive index at given wavelength.

        Args:
            wl (ArrayLike): Wavelength [μm].
        """
        _wl = np.asarray(wl)
        self.bound_check(_wl, "n")
        return self.__n(_wl)

    def k(self, wl: ArrayLike) -> np.ndarray:
        """Return extinction coefficient at given wavelength.

        Args:
            wl (ArrayLike): Wavelength [μm].
        """
        _wl = np.asarray(wl)
        self.bound_check(_wl, "k")
        return self.__k(_wl)

    def eps(self, wl: ArrayLike) -> np.ndarray:
        """Return complex dielectric constant at given wavelength.

        Args:
            wl (Union[float, complex, Sequence, np.ndarray]): Wavelength [μm].
        """
        _wl = np.asarray(wl)
        if self.f > 20:
            self.bound_check(_wl, "nk")
            return self.formula(_wl)
        n: np.ndarray = self.n(_wl)
        k: np.ndarray = self.k(_wl)
        eps = n ** 2 - k ** 2 + 2j * n * k
        return eps


class ConstMaterial(AbstractMaterial):
    """A class defines a material with constant permittivity

    Attributes:
        ce (complex): The value of constant permittivity
        label (str): A label used in plot
    """

    def __init__(self, params: dict) -> None:
        """Initialize Material

        Args:
            params (Dict): parameter dict contains the following key and values
                'RI': Constant refractive index. (complex)
                'e': Constant permittivity. (complex)
        """
        if "RI" in params:
            RI = params["RI"]
            self.ce = RI ** 2 + 0.0j
            self.cn = RI.real
            self.ck = RI.imag
            self.label = f"RI: {RI}"
            if "e" in params:
                e = params["e"]
                if e != RI ** 2:
                    raise ValueError("e must be RI ** 2.")
        elif "e" in params:
            e = params["e"]
            self.label = f"eps: {e}"
            self.ce = e + 0.0j
            ri = np.sqrt(_ensure_positive_imag(e))
            self.cn = ri.real
            self.ck = ri.imag
        else:
            raise ValueError("'RI' or 'e' must be specified")

    def n(self, wl: ArrayLike) -> np.ndarray:
        """Return refractive index at given wavelength.

        Args:
            wl (ArrayLike): Wavelength [μm].
        """
        return self.cn * np.ones_like(wl)

    def k(self, wl: ArrayLike) -> np.ndarray:
        """Return extinction coefficient at given wavelength.

        Args:
            wl (ArrayLike): Wavelength [μm].
        """
        return self.ck * np.ones_like(wl)

    def eps(self, wl: ArrayLike) -> np.ndarray:
        """Return complex dielectric constant at given wavelength.

        Args:
            wl (Union[float, complex, Sequence, np.ndarray]): Wavelength [μm].
        """
        return self.ce * np.ones_like(wl)


class PEC(ConstMaterial):
    """Perfect Electric Conductor class as a material that has negative large pemittivity

    Attributes:
        ce (-1e8+0j: complex): The value of constant permittivity
        label ('PEC': str): A label used in plot
    """

    def __init__(
        self,
    ) -> None:
        """Initialize Material

        Args:
            rid (RiiDataFrame): Rii_Pandas DataFrame.
        """
        self.label = "PEC"
        self.ce = -1.0e8 + 0.0j
        self.cn = 0.0
        self.ck = 1.0e4


class Material(AbstractMaterial):
    """A Class that constructs RiiMaterial or ConstMaterial instance depending on the given parameters.

    Implement __call__ method that calculate the permittivity at a single value of angular frequency.
    Introduce 'im_factor' that is a magnification factor multiplied to the imaginary part of permittivity.

    Args:
        AbstractMaterial ([type]): [description]
    """

    def __init__(
        self, params: dict, rid: Optional[riip.dataframe.RiiDataFrame] = None
    ) -> None:
        """Initialize Material

        Args:
            params (dict): parameter dict contains the following key and values
                'PEC' (bool): If the target material is PEC. Deafaults to False.
                'id' (int): ID number.
                'book' (str): book value in catalog of RiiDataFrame.
                'page' (str): page value in catalog of RiiDataFrame.
                'RI' (complex): Constant refractive index.
                'e' (complex): Constant permittivity.
                'bound_check' (bool): True if bound check should be done. Defaults to True.
                'im_factor' (float): A magnification factor multiplied to the imaginary part of permittivity. Defaults to 1.0.
            rid (RiiDataFrame): Rii_Pandas DataFrame. Defaults to None.
        """
        self.params = params
        if params.get("PEC", False):
            self.material: PEC | ConstMaterial | RiiMaterial = PEC()
            self.__ce0: Optional[complex] = self.material.ce
            self.f = 0
        elif "RI" in params or "e" in params:
            self.material = ConstMaterial(params)
            self.__ce0 = self.material.ce
            self.f = 0
        elif "id" not in params and ("book" not in params or "page" not in params):
            raise ValueError(
                "'PEC', 'RI', 'e', 'id', or 'book'-'page' pair must be specified"
            )
        else:
            if rid is None:
                rid = riip.dataframe.RiiDataFrame()
            if "book" in params and "page" in params:
                idx = rid.book_page_to_id(params)
                if "id" in "params":
                    idx != params["id"]
                    raise ValueError(
                        "There is an inconsistency between 'id' and 'book'-'page' pair"
                    )
                else:
                    params["id"] = idx
            self.material = RiiMaterial(
                params["id"], rid.catalog, rid.raw_data, params.get("bound_check", True)
            )
            self.catalog = self.material.catalog
            self.wl_max = self.catalog["wl_max"]
            self.wl_min = self.catalog["wl_min"]
            self.raw_data = self.material.raw_data
            self.bound_check_flag = self.material.bound_check_flag
            self.__ce0 = None
            self.f = self.material.f
            if self.f != 0:
                self.cs = self.material.cs
        self.__w: Optional[float | complex] = None
        self.__ce: Optional[complex] = self.__ce0
        self.im_factor = params.get("im_factor", 1.0)

    def eps(self, wl: ArrayLike) -> np.ndarray:
        """Return complex dielectric constant at given wavelength.

        Args:
            wl (Union[float, complex, Sequence, np.ndarray]): Wavelength [μm].
        """
        e = self.material.eps(wl)
        if self.__im_factor != 1.0:
            imag = e.imag * self.__im_factor
            e = e.real + 1j * imag
        return e

    def n(self, wl: ArrayLike) -> np.ndarray:
        """Return refractive index at given wavelength.

        Args:
            wl (ArrayLike): Wavelength [μm].
        """
        return np.sqrt(_ensure_positive_imag(self.eps(wl))).real

    def k(self, wl: ArrayLike) -> np.ndarray:
        """Return extinction coefficient at given wavelength.

        Args:
            wl (ArrayLike): Wavelength [μm].
        """
        return np.sqrt(_ensure_positive_imag(self.eps(wl))).imag

    @property
    def im_factor(self) -> float:
        return self.__im_factor

    @im_factor.setter
    def im_factor(self, factor: float) -> None:
        self.__w = None
        self.__im_factor = factor
        if factor != 1.0:
            self.label = self.material.label + f" im_factor: {factor}"
            if self.__ce0 is not None:
                if factor != 1.0:
                    imag = self.__ce0 * factor
                    self.__ce = self.__ce0.real + 1j * imag
        else:
            self.label = self.material.label

    def __call__(self, w: float | complex) -> complex:
        """Return relative permittivity at given angular frequency.

        Args:
            w (float | complex): A float indicating the angular frequency (vacuum wavenumber ω/c [rad/μm]).

        Returns:
            complex: Relative permittivity at w
        """
        if self.__ce is not None:
            return self.__ce
        if self.__w is None or w != self.__w:
            wl = 2 * np.pi / w.real
            if self.bound_check_flag and (wl < self.wl_min or wl > self.wl_max):
                raise ValueError(
                    f"Wavelength {wl} is out of bounds [{self.wl_min} {self.wl_max}][um]"
                )
            if self.f > 0:
                if self.f > 20:
                    self.__e = formulas_cython_dict[self.f](w, self.cs)
                else:
                    _n = formulas_cython_dict[self.f](w.real, self.cs)
                    _k = self.material.k(2 * np.pi / w.real).item()
                    self.__e = _n ** 2 - _k ** 2 + 2j * _n * _k
                if self.__im_factor != 1.0:
                    imag = self.__e.imag * self.__im_factor
                    self.__e = self.__e.real + 1j * imag
            else:
                self.__e = self.eps(wl).item()
            self.__w = w
        return self.__e



from __future__ import annotations

from collections.abc import Callable
from itertools import islice

import numpy as np
from scipy import constants
from scipy.special import wofz

from .formulas_cython import (
    formula_1_cython,
    formula_2_cython,
    formula_3_cython,
    formula_4_cython,
    formula_5_cython,
    formula_6_cython,
    formula_7_cython,
    formula_8_cython,
    formula_9_cython,
    formula_21_cython,
    formula_22_cython,
)

UNIT = constants.h * constants.c * 1e6 / constants.e


def _formula_1(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    x_sqr: np.ndarray = x ** 2
    n_sqr = 1 + cs[0]
    c1 = 0.0
    c2 = 0.0
    for c1, c2 in zip(islice(cs, 1, None, 2), islice(cs, 2, None, 2)):
        if c1 == 0.0:
            break
        n_sqr += c1 * x_sqr / (x_sqr - c2 ** 2)
    return np.sqrt(n_sqr * (n_sqr > 0))


def _formula_2(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    x_sqr = x ** 2
    n_sqr: float = 1 + cs[0]
    c1: float = 0.0
    c2: float = 0.0
    for c1, c2 in zip(islice(cs, 1, None, 2), islice(cs, 2, None, 2)):
        if c1 == 0.0:
            break
        n_sqr += c1 * x_sqr / (x_sqr - c2)
    return np.sqrt(n_sqr * (n_sqr > 0))


def _formula_3(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    n_sqr: float = cs[0]
    c1 = 0.0
    c2 = 0.0
    for c1, c2 in zip(islice(cs, 1, None, 2), islice(cs, 2, None, 2)):
        if c1 == 0.0:
            break
        n_sqr += c1 * x ** c2
    return np.sqrt(n_sqr * (n_sqr > 0))


def _formula_4(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    n_sqr = (
        cs[0]
        + cs[1] * x ** cs[2] / (x ** 2 - cs[3] ** cs[4])
        + cs[5] * x ** cs[6] / (x ** 2 - cs[7] ** cs[8])
    )
    c1 = 0.0
    c2 = 0.0
    for c1, c2 in zip(islice(cs, 9, None, 2), islice(cs, 10, None, 2)):
        if c1 == 0.0:
            break
        n_sqr += c1 * x ** c2
    return np.sqrt(n_sqr * (n_sqr > 0))


def _formula_5(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    n = cs[0]
    c1 = 0.0
    c2 = 0.0
    for c1, c2 in zip(islice(cs, 1, None, 2), islice(cs, 2, None, 2)):
        if c1 == 0.0:
            break
        n += c1 * x ** c2
    return n


def _formula_6(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    x_m2 = 1 / x ** 2
    n = 1 + cs[0]
    c1 = 0.0
    c2 = 0.0
    for c1, c2 in zip(islice(cs, 1, None, 2), islice(cs, 2, None, 2)):
        if c1 == 0.0:
            break
        n += c1 / (c2 - x_m2)
    return n


def _formula_7(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    x_sqr = x ** 2
    n = (
        cs[0]
        + cs[1] / (x_sqr - 0.028)
        + cs[2] / (x_sqr - 0.028) ** 2
        + cs[3] * x_sqr
        + cs[4] * x_sqr ** 2
        + cs[5] * x_sqr ** 3
    )
    return n


def _formula_8(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    x_sqr = x ** 2
    a = cs[0] + cs[1] * x_sqr / (x_sqr - cs[2]) + cs[3] * x_sqr
    n_sqr = (1 + 2 * a) / (1 - a)
    return np.sqrt(n_sqr * (n_sqr > 0))


def _formula_9(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    n_sqr = (
        cs[0]
        + cs[1] / (x ** 2 - cs[2])
        + cs[3] * (x - cs[4]) / ((x - cs[4]) ** 2 + cs[5])
    )
    return np.sqrt(n_sqr * (n_sqr > 0))


def _formula_21(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    w = UNIT / x
    eb = cs[0]
    f0 = cs[1]
    g0 = cs[2]
    wp = cs[3]
    eps = eb - f0 * wp ** 2 / (w ** 2 + 1j * w * g0)
    for fj, gj, wj in zip(
        islice(cs, 4, None, 3), islice(cs, 5, None, 3), islice(cs, 6, None, 3)
    ):
        if fj == 0:
            break
        eps -= fj * wp ** 2 / (w ** 2 - wj ** 2 + 1j * w * gj)
    return eps


def _formula_22(x: np.ndarray, cs: np.ndarray) -> np.ndarray:
    w = UNIT / x
    eb = cs[0]
    f0 = cs[1]
    g0 = cs[2]
    wp = cs[3]
    c = 1j * np.sqrt(np.pi) / (2 * np.sqrt(2))
    eps = eb - f0 * wp ** 2 / (w ** 2 + 1j * w * g0)
    for fj, gj, wj, sj in zip(
        islice(cs, 4, None, 4),
        islice(cs, 5, None, 4),
        islice(cs, 6, None, 4),
        islice(cs, 7, None, 4),
    ):
        if fj == 0:
            break
        aj = np.sqrt(w * (w + 1j * gj))
        sj_inv = 1 / sj if sj != 0 else 0
        eps += (
            c
            * fj
            * wp ** 2
            / aj
            * sj_inv
            * (
                wofz((aj - wj) / np.sqrt(2) * sj_inv)
                + wofz((aj + wj) / np.sqrt(2) * sj_inv)
            )
        )
    return eps


formulas_numpy_dict: dict[int, Callable] = {
    1: _formula_1,
    2: _formula_2,
    3: _formula_3,
    4: _formula_4,
    5: _formula_5,
    6: _formula_6,
    7: _formula_7,
    8: _formula_8,
    9: _formula_9,
    21: _formula_21,
    22: _formula_22,
}

formulas_cython_dict: dict[int, Callable] = {
    1: formula_1_cython,
    2: formula_2_cython,
    3: formula_3_cython,
    4: formula_4_cython,
    5: formula_5_cython,
    6: formula_6_cython,
    7: formula_7_cython,
    8: formula_8_cython,
    9: formula_9_cython,
    21: formula_21_cython,
    22: formula_22_cython,
}



from __future__ import annotations

import os
from collections import OrderedDict
from collections.abc import Iterable, Sequence
from logging import ERROR, WARNING, getLogger
from typing import ClassVar, Optional

import git
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import yaml
from IPython.display import HTML
from pandas import DataFrame
from pandas.core.series import Series

import riip.material

# from numpy.typing import ArrayLike


logger = getLogger(__package__)
_dirname = os.path.dirname(__file__)
_ri_database = os.path.join(_dirname, "data", "refractiveindex.info-database")
_db_directory = os.path.join(_ri_database, "database")
_my_db_directory = os.path.join(_dirname, "data", "my_database")
_catalog_file = os.path.join(_dirname, "data", "catalog.csv")
_raw_data_file = os.path.join(_dirname, "data", "raw_data.csv")
_grid_data_file = os.path.join(_dirname, "data", "grid_data.h5")
_ri_database_repo = (
    "https://github.com/polyanskiy/" + "refractiveindex.info-database.git"
)
_ri_database_patch = os.path.join(_dirname, "data", "riid.patch")


class RiiDataFrame:
    """A class that provides a Pandas DataFrame for 'refractiveindex.info database'.

    Attributes:
        db_path: The path to the refractiveindex.info-database/database.
        my_db_path: The path to my_database.
        catalog: The catalog.
        catalog_file: The csv filename to store the catalog.
        raw_data: The experimental data.
        raw_data_file: The csv filename to store the raw_data.
        grid_data_file: The hdf5 filename to store the grid_data.

    """

    _catalog_columns: ClassVar[OrderedDict] = OrderedDict(
        (
            ("id", np.int32),
            ("shelf", str),
            ("shelf_name", str),
            ("division", str),
            ("book", str),
            ("book_name", str),
            ("section", str),
            ("page", str),
            ("path", str),
            ("formula", np.int32),
            ("tabulated", str),
            ("num_n", np.int32),
            ("num_k", np.int32),
            ("wl_n_min", np.float64),
            ("wl_n_max", np.float64),
            ("wl_k_min", np.float64),
            ("wl_k_max", np.float64),
            ("wl_min", np.float64),
            ("wl_max", np.float64),
        )
    )

    _raw_data_columns: ClassVar[OrderedDict] = OrderedDict(
        (
            ("id", np.int32),
            ("c", np.float64),
            ("wl_n", np.float64),
            ("n", np.float64),
            ("wl_k", np.float64),
            ("k", np.float64),
        )
    )

    _grid_data_columns: ClassVar[OrderedDict] = OrderedDict(
        (("id", np.int32), ("wl", np.float64), ("n", np.float64), ("k", np.float64))
    )

    def __init__(
        self,
        db_path: str = _db_directory,
        catalog_file: str = _catalog_file,
        raw_data_file: str = _raw_data_file,
        grid_data_file: str = _grid_data_file,
        my_db_path: str = _my_db_directory,
    ):
        """Initialize the RiiDataFrame.

        Args:
            db_path: The path to the refractiveindex.info-database/database.
            my_db_path: The path to my_database.
            catalog_file: The filename of the catalog csv file.
            raw_data_file: The filename of the experimental data csv file.
            grid_data_file: The filename of the grid wl-nk data csv file.
        """
        self._db_path: str = db_path
        self._my_db_path: str = my_db_path
        self._ri_database: str = os.path.dirname(self._db_path)
        self._catalog_file: str = catalog_file
        self._raw_data_file: str = raw_data_file
        self._grid_data_file: str = grid_data_file
        _catalog, _raw_data = self._load_catalog_and_raw_data()
        self.catalog: DataFrame = _catalog
        self.raw_data: DataFrame = _raw_data
        self.__book_page_order = self._create_book_page_order()

    def _load_catalog_and_raw_data(self) -> tuple[DataFrame, DataFrame]:
        # Create csv files
        if not os.path.isfile(self._catalog_file):
            logger.warning("Catalog file not found.")
            if not os.path.isfile(os.path.join(self._db_path, "library.yml")):
                logger.warning("Cloning Repository...")
                repo = git.Repo.clone_from(
                    _ri_database_repo, self._ri_database, branch="master"
                )
                repo.git.apply(_ri_database_patch)
                # git.Repo.clone_from(
                #     _ri_database_repo, self._ri_database, branch="master"
                # )
                logger.warning("Done.")
            logger.warning("Creating catalog file...")
            catalog = self._add_my_db_to_catalog(self._create_catalog())
            logger.warning("Done.")

            # Preparing raw_data
            logger.warning("Creating raw data file...")
            raw_data, catalog = self._create_raw_data_and_modify_catalog(catalog)
            logger.warning("Done.")

            # Preparing grid_data
            logger.warning("Updating grid data file...")
            catalog = catalog.set_index("id")
            raw_data = raw_data.set_index("id")
            self._create_grid_data(catalog, raw_data)
            logger.warning("Done.")
        else:
            catalog = pd.read_csv(
                self._catalog_file,
                dtype=self._catalog_columns,
                index_col="id",
                na_filter=False,
            )
            raw_data = pd.read_csv(
                self._raw_data_file,
                dtype=self._raw_data_columns,
                index_col="id",
                na_filter=False,
            )
        return catalog, raw_data

    @staticmethod
    def _extract_entry(db_path: str, start_id: int = 0) -> Iterable:
        """Yield a single data set."""
        reference_path = os.path.normpath(db_path)
        library_file = os.path.join(reference_path, "library.yml")
        with open(library_file, "r", encoding="utf-8") as f:
            library = yaml.safe_load(f)
        idx = start_id
        shelf = "main"
        book = "Ag (Experimental data)"
        page = "Johnson"
        try:
            for sh in library:
                shelf = sh["SHELF"]
                if shelf == "3d":
                    # This shelf does not seem to contain new data.
                    break
                shelf_name = sh["name"]
                division = None
                for b in sh["content"]:
                    if "DIVIDER" in b:
                        division = b["DIVIDER"]
                    else:
                        if division is None:
                            raise Exception("'DIVIDER' is missing in 'library.yml'.")
                        if "DIVIDER" not in b["content"]:
                            section = ""
                        for p in b["content"]:
                            if "DIVIDER" in p:
                                # This DIVIDER specifies the phase of the
                                #  material such as gas, liquid or solid, so it
                                #  is added to the book and book_name with
                                #  parentheses.
                                section = p["DIVIDER"]
                            else:
                                book = b["BOOK"]
                                book_name = b["name"]
                                page = p["PAGE"]
                                path = os.path.join(
                                    reference_path, "data", os.path.normpath(p["data"])
                                )
                                logger.debug("{0} {1} {2}".format(idx, book, page))
                                row = [idx, shelf, shelf_name, division]
                                row += [book, book_name, section, page, path]
                                row += [0, "f", 0, 0, 0, 0, 0, 0, 0, 0]
                                yield row
                                idx += 1
        except Exception as e:
            message = (
                "There seems to be some inconsistency in the library.yml "
                + "around id={}, shelf={}, book={}, page={}.".format(
                    idx, shelf, book, page
                )
            )
            raise Exception(message) from e

    def _create_catalog(self) -> DataFrame:
        """Create catalog DataFrame from library.yml."""
        logger.info("Creating catalog...")
        df = DataFrame(
            list(self._extract_entry(self._db_path)),
            columns=self._catalog_columns.keys(),
        )
        logger.info("Done.")
        return df.astype(self._catalog_columns)

    def _add_my_db_to_catalog(self, catalog: DataFrame) -> DataFrame:
        """Add data in my_database to catalog DataFrame."""
        logger.info("Adding my_db to catalog...")
        start_id = catalog["id"].values[-1] + 1
        logger.debug(start_id)
        df = DataFrame(
            list(self._extract_entry(self._my_db_path, start_id)),
            columns=self._catalog_columns.keys(),
        )
        df = catalog.append(df, ignore_index=True)
        logger.info("Done.")
        return df

    def _create_book_page_order(self) -> Series:
        """Create [id, book+page string] array used to search id."""
        cl = self.catalog
        book_page = {
            idx: f"{cl.loc[idx, 'book']}{cl.loc[idx, 'page']}" for idx in cl.index
        }
        return Series(book_page).sort_values()

    def book_page_to_id(self, params: dict) -> int:
        bp = params["book"] + params["page"]
        ind = np.searchsorted(self.__book_page_order, bp)
        return self.__book_page_order.index[ind]

    def _extract_raw_data(
        self, idx: int, catalog: DataFrame
    ) -> tuple[DataFrame, DataFrame]:
        """Yield a single raw data set.

        Some data are inserted into the catalog.

        Args:
            catalog: The catalog.
            idx: The ID number of the data set.
        """
        path = catalog.loc[idx, "path"]
        with open(path, "r", encoding="utf-8") as f:
            data_list = yaml.safe_load(f)["DATA"]
        wl_n_min = wl_k_min = 0.0
        wl_n_max = wl_k_max = np.inf
        formula = 0
        tabulated = ""
        cs = []
        wls_n = []
        wls_k = []
        ns = []
        ks = []
        num_n = num_k = 0
        for data in data_list:
            data_type, data_set = data["type"].strip().split()

            # For tabulated data
            if data_type == "tabulated":
                if data_set == "nk":
                    tabulated += data_set
                    wls_n, ns, ks = np.array(
                        [
                            line.strip().split()
                            for line in data["data"].strip().split("\n")
                        ],
                        dtype=float,
                    ).T
                    wls_n, inds = np.unique(wls_n, return_index=True)
                    ns = ns[inds]
                    ks = ks[inds]
                    inds = np.argsort(wls_n)
                    wls_n = list(wls_n[inds])
                    wls_k = wls_n
                    ns = list(ns[inds])
                    ks = list(ks[inds])
                    wl_n_min = wl_k_min = wls_n[0]
                    wl_n_max = wl_k_max = wls_n[-1]
                    num_n = len(wls_n)
                    num_k = len(wls_k)
                elif data_set == "n":
                    tabulated += data_set
                    wls_n, ns = np.array(
                        [
                            line.strip().split()
                            for line in data["data"].strip().split("\n")
                        ],
                        dtype=float,
                    ).T
                    wls_n, inds = np.unique(wls_n, return_index=True)
                    ns = ns[inds]
                    inds = np.argsort(wls_n)
                    wls_n = list(wls_n[inds])
                    ns = list(ns[inds])
                    wl_n_min = wls_n[0]
                    wl_n_max = wls_n[-1]
                    num_n = len(wls_n)
                elif data_set == "k":
                    tabulated += data_set
                    wls_k, ks = np.array(
                        [
                            line.strip().split()
                            for line in data["data"].strip().split("\n")
                        ],
                        dtype=float,
                    ).T
                    wls_k, inds = np.unique(wls_k, return_index=True)
                    ks = ks[inds]
                    inds = np.argsort(wls_k)
                    wls_k = list(wls_k[inds])
                    ks = list(ks[inds])
                    wl_k_min = wls_k[0]
                    wl_k_max = wls_k[-1]
                    num_k = len(wls_k)
                else:
                    raise Exception("DATA is broken.")
            # For formulas
            elif data_type == "formula":
                formula = data_set
                wl_n_min, wl_n_max = [
                    float(s) for s in data["wavelength_range"].strip().split()
                ]
                cs = [float(s) for s in data["coefficients"].strip().split()]
            else:
                raise Exception("DATA has unknown contents {}".format(data_type))

        if len(tabulated) > 2:
            raise Exception("Too many tabulated data set are provided")
        elif "nn" in tabulated or "kk" in tabulated:
            raise Exception("There is redundancy in n or k.")
        elif tabulated == "kn":
            tabulated = "nk"
        elif tabulated == "":
            tabulated = "f"

        if tabulated == "k" and formula != 0:
            if wl_n_min < wl_k_min:
                wls_k = [wl_n_min] + wls_k
                ks = [min(ks)] + ks
                num_k += 1
            if wl_n_max > wl_k_max:
                wls_k = wls_k + [wl_n_max]
                ks = ks + [min(ks)]
                num_k += 1
            wl_k_min, wl_k_max = wl_n_min, wl_n_max

        if "k" not in tabulated:
            wl_k_min, wl_k_max = wl_n_min, wl_n_max

        wl_min = max(wl_n_min, wl_k_min)
        wl_max = min(wl_n_max, wl_k_max)

        # The coefficients not included in the formula must be zero.
        num_c = len(cs)
        if formula != 0:
            cs += [0.0] * (24 - num_c)
            num_c = 24

        # All the arrays must have the same length.
        num = max(num_n, num_k, num_c)
        _cs = np.array(cs + [0.0] * (num - num_c), dtype=np.float64)
        _wls_n = np.array(wls_n + [0.0] * (num - num_n), dtype=np.float64)
        _ns = np.array(ns + [0.0] * (num - num_n), dtype=np.float64)
        _wls_k = np.array(wls_k + [0.0] * (num - num_k), dtype=np.float64)
        _ks = np.array(ks + [0.0] * (num - num_k), dtype=np.float64)

        # Rewrite catalog with the obtained data
        catalog.loc[idx, "formula"] = formula
        catalog.loc[idx, "tabulated"] = tabulated
        catalog.loc[idx, "num_n"] = num_n
        catalog.loc[idx, "num_k"] = num_k
        catalog.loc[idx, "wl_n_min"] = wl_n_min
        catalog.loc[idx, "wl_n_max"] = wl_n_max
        catalog.loc[idx, "wl_k_min"] = wl_k_min
        catalog.loc[idx, "wl_k_max"] = wl_k_max
        catalog.loc[idx, "wl_min"] = wl_min
        catalog.loc[idx, "wl_max"] = wl_max

        df = DataFrame(
            {
                key: val
                for key, val in zip(
                    self._raw_data_columns.keys(), [idx, _cs, _wls_n, _ns, _wls_k, _ks]
                )
            }
        )
        # Arrange the columns according to the order of _raw_data_columns
        df = df.loc[:, self._raw_data_columns.keys()].astype(self._raw_data_columns)
        return df, catalog

    def _create_raw_data_and_modify_catalog(
        self, catalog: DataFrame
    ) -> tuple[DataFrame, DataFrame]:
        """Create a DataFrame for experimental data."""
        logger.info("Creating raw data...")
        df = DataFrame(columns=self._raw_data_columns)
        for idx in catalog.index:
            logger.debug("{}: {}".format(idx, catalog.loc[idx, "path"]))
            df_idx, catalog = self._extract_raw_data(idx, catalog)
            df = df.append(df_idx, ignore_index=True)
        df = df.astype(self._raw_data_columns)
        catalog.to_csv(self._catalog_file, index=False, encoding="utf-8")
        df.to_csv(self._raw_data_file, index=False, encoding="utf-8")
        logger.info("Done.")
        return df, catalog

    def load_grid_data(self, id: Optional[int] = None) -> DataFrame:
        """Load grid data of (wl, n, k) for given id.

        Args:
            id (Optional[int]): ID number. If id is None, all the data is loaded.
                Defaults to None.

        Returns:
            DataFrame: Grid data of (wl, n, k).
                (wl, n, k) = (wavelength, refractive index, extinction coefficient).
        """
        if not os.path.isfile(self._grid_data_file):
            logger.warning("Grid data file not found.")
            logger.warning("Creating grid data file...")
            self._create_grid_data(self.catalog, self.raw_data)
            logger.warning("Done.")
        else:
            logger.info("Grid data file found at {}".format(self._grid_data_file))
        if id is None:
            return pd.read_hdf(self._grid_data_file).set_index("id")
        return pd.read_hdf(self._grid_data_file, where=f"id == {id}").set_index("id")

    def _create_grid_data(self, catalog: DataFrame, raw_data: DataFrame) -> None:
        """Create a DataFrame for the wl-nk data."""
        logger.info("Creating grid data...")
        columns = self._grid_data_columns.keys()
        df = DataFrame(columns=columns)
        logger.setLevel(ERROR)
        for idx in catalog.index:
            material = riip.material.RiiMaterial(idx, catalog, raw_data)
            wl_min = material.catalog.loc["wl_min"]
            wl_max = material.catalog.loc["wl_max"]
            wls = np.linspace(wl_min, wl_max, 200)
            ns = material.n(wls)
            ks = material.k(wls)
            data = {key: val for key, val in zip(columns, [idx, wls, ns, ks])}
            df = df.append(DataFrame(data).loc[:, columns], ignore_index=True)
        logger.setLevel(WARNING)
        df = df.astype(self._grid_data_columns)
        df.to_hdf(
            self._grid_data_file,
            "grid_data",
            mode="w",
            data_columns=["id"],
            format="table",
        )
        logger.info("Done.")

    def update_db(self) -> None:
        """Pull repository and update local database."""
        if not os.path.isfile(os.path.join(self._db_path, "library.yml")):
            logger.warning("Cloning Repository.")
            git.Repo.clone_from(_ri_database_repo, self._ri_database, branch="master")
            logger.warning("Done.")
        else:
            logger.warning("Pulling Repository...")
            repo = git.Repo(self._ri_database)
            repo.remotes.origin.pull()
            logger.warning("Done.")
        logger.warning("Updating catalog file...")
        catalog = self._add_my_db_to_catalog(self._create_catalog())
        logger.warning("Done.")
        logger.warning("Updating raw data file...")
        self.raw_data, self.catalog = self._create_raw_data_and_modify_catalog(catalog)
        self.catalog = self.catalog.set_index("id")
        self.raw_data = self.raw_data.set_index("id")
        logger.warning("Done.")
        logger.warning("Updating grid data file...")
        self._create_grid_data(self.catalog, self.raw_data)
        logger.warning("Done.")
        logger.warning("All Done.")

        """."""

    def search(self, name: str) -> DataFrame:
        """Search pages which contain the name of material.

        Args:
            name (str): Name of material

        Returns:
            DataFrame: Simplified catalog
        """
        columns = [
            "book",
            "section",
            "page",
            "formula",
            "tabulated",
            "wl_min",
            "wl_max",
        ]
        df = self.catalog[
            (
                (self.catalog["book"].str.contains(name))
                | (
                    self.catalog["book_name"]
                    .str.replace("<sub>", "")
                    .str.replace("</sub>", "")
                    .str.lower()
                    .str.contains(name.lower())
                )
            )
        ]
        return df.loc[:, columns]

    def select(self, cond: str) -> DataFrame:
        """Select pages that fullfil the condition.

        Args:
            cond (str): Query condition, such as '1.5 <= n <= 2 & 1.0 <= wl <= 2.0'

        Returns:
            List[int]: Simplified catalog
        """
        columns = [
            "book",
            "section",
            "page",
            "formula",
            "tabulated",
            "wl_min",
            "wl_max",
        ]
        gd = self.load_grid_data()
        id_list = gd.query(cond).index.unique()
        return self.catalog.loc[id_list, columns]

    def show(self, id: int | Sequence[int]) -> DataFrame:
        """Summary of page(s) of ID (list of IDs).

        Args:
            id (Union[int, Sequence[int]]): ID number

        Returns:
            DataFrame: Simplified catalog
        """
        columns = [
            "book",
            "section",
            "page",
            "formula",
            "tabulated",
            "wl_min",
            "wl_max",
        ]
        return self.catalog.loc[id, columns]

    def material(self, params: dict) -> riip.material.Material:
        """Create instance of Material class associated with ID.

        Args:
            params (dict): Parameter dict that can contain the following values:
                'id': ID number (int)
                'book': book value in catalog of RiiDataFrame. (str)
                'page': page value in catalog of RiiDataFrame. (str)
                'RI': Constant refractive index. (complex)
                'e': Constant permittivity. (complex)
                'bound_check': True if bound check should be done. Defaults to True. (bool)
                'im_factor': A magnification factor multiplied to the imaginary part of permittivity. Defaults to 1.0. (float)

        Returns:
            Material: A class that provides refractive index, extinction coefficient and dielectric function of the material
        """
        return riip.material.Material(params, self)

    def read(self, id: int, as_dict: bool = False):
        """Return contants of a page associated with the id.

        Args:
            id (int): ID number
            as_dict (bool): If True, the page contents are returned as python dict
        Returns:
            Union[str, dict]: Page contents
        """
        path = self.catalog.loc[id, "path"]
        with open(path) as fd:
            if as_dict:
                contents = yaml.safe_load(fd)
            else:
                contents = fd.read()
        return contents

    def references(self, id: int) -> HTML:
        """Return REFERENCES as IPython.display.HTML class.

        Args:
            id (int): ID number

        Returns:
            HTML: REFERENCES as IPython.display.HTML class
        """
        contents = self.read(id, as_dict=True)
        return HTML(contents["REFERENCES"])

    def plot(
        self,
        id: int,
        comp: str = "n",
        fmt1: Optional[str] = "-",
        fmt2: Optional[str] = "--",
        **kwargs,
    ):
        """Plot refractive index, extinction coefficient or permittivity.

        Args:
            id (int): ID number.
            comp (str): 'n', 'k' or 'eps'.
            fmt1 (Union[str, None]): Plot format for n and Re(eps).
            fmt2 (Union[str, None]): Plot format for k and Im(eps).
        """
        label = f"{self.catalog.loc[id, 'book']} {self.catalog.loc[id, 'page']}"
        df = self.load_grid_data(id)
        wls = df["wl"]
        ns = df["n"]
        ks = df["k"]
        kwargs.setdefault("lw", 4)
        kwargs.setdefault("ms", 8)
        if comp == "n":
            plt.plot(wls, ns, fmt1, label=label, **kwargs)
            plt.ylabel(r"$n$")
        elif comp == "k":
            plt.plot(wls, ks, fmt2, label=label, **kwargs)
            plt.ylabel(r"$k$")
        elif comp == "eps":
            eps_r = ns ** 2 - ks ** 2
            eps_i = 2 * ns * ks
            (line,) = plt.plot(wls, eps_r, fmt1, label=label, **kwargs)
            color = line.get_color()
            plt.plot(wls, eps_i, fmt2, color=color, **kwargs)
            plt.ylabel(r"$\varepsilon$")
        plt.xlabel(r"$\lambda$ $[\mathrm{\mu m}]$")
        plt.legend()


if __name__ == "__main__":
    from logging import DEBUG, Formatter, StreamHandler, getLogger

    logger = getLogger("")
    formatter = Formatter(fmt="%(levelname)s:[%(name)s.%(funcName)s]: %(message)s")
    logger.setLevel(DEBUG)
    stream_handler = StreamHandler()
    stream_handler.setFormatter(formatter)
    stream_handler.setLevel(DEBUG)
    logger.addHandler(stream_handler)

    rii_df = RiiDataFrame()
    rii_df.update_db()



from .dataframe import RiiDataFrame
from .material import Material

# __all__ = [
#     "Material",
#     "RiiDataFrame",
# ]

__author__ = "Munehiro Nishida"
__version__ = "0.6.1"
__license__ = "MIT"



import numpy as np
from scipy import constants

cimport cython
cimport numpy as np
from libc.math cimport M_PI, sqrt
from scipy.special.cython_special cimport wofz

ctypedef np.complex128_t cdouble

cdef extern from "<complex>" namespace "std" nogil:
    cdouble csqrt "sqrt" (cdouble z)
    double creal "real" (cdouble z)

cdef:
    double UNIT = constants.h * constants.c * 1e6 / constants.e / (2 * M_PI)
    cdouble C_I = 1.0j
    cdouble C = C_I * sqrt(M_PI) / 2
    double RT2_INV = 1 / sqrt(2)


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_1_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        double x_sqr = (2 * M_PI / creal(w)) ** 2
        double n_sqr = 1 + cs[0]
        double c1, c2

    for i in range(11):
        c1 = cs[2 * i + 1]
        c2 = cs[2 * i + 2]
        if c1 == 0:
            break
        n_sqr += c1 * x_sqr / (x_sqr - c2 ** 2)
    return sqrt(n_sqr * (n_sqr > 0))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_2_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        double x_sqr = (2 * M_PI / creal(w)) ** 2
        double n_sqr = 1 + cs[0]
        double c1, c2
    for i in range(11):
        c1 = cs[2 * i + 1]
        c2 = cs[2 * i + 2]
        if c1 == 0:
            break
        n_sqr += c1 * x_sqr / (x_sqr - c2)
    return sqrt(n_sqr * (n_sqr > 0))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_3_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        double x = 2 * M_PI / creal(w)
        double n_sqr = cs[0]
        double c1, c2
    for i in range(11):
        c1 = cs[2 * i + 1]
        c2 = cs[2 * i + 2]
        if c1 == 0:
            break
        n_sqr += c1 * x ** c2
    return sqrt(n_sqr * (n_sqr > 0))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_4_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        double x = 2 * M_PI / creal(w)
        double n_sqr = (
            cs[0]
            + cs[1] * x ** cs[2] / (x ** 2 - cs[3] ** cs[4])
            + cs[5] * x ** cs[6] / (x ** 2 - cs[7] ** cs[8])
        )
    for i in range(7):
        c1 = cs[2 * i + 9]
        c2 = cs[2 * i + 10]
        if c1 == 0:
            break
        n_sqr += c1 * x ** c2
    return sqrt(n_sqr * (n_sqr > 0))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_5_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        double n = cs[0]
        double x = 2 * M_PI / creal(w)
    for i in range(11):
        c1 = cs[2 * i + 1]
        c2 = cs[2 * i + 2]
        if c1 == 0:
            break
        n += c1 * x ** c2
    return n


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_6_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        double x_m2 = (creal(w) / (2 * M_PI)) ** 2
        double n = 1 + cs[0]
    for i in range(11):
        c1 = cs[2 * i + 1]
        c2 = cs[2 * i + 2]
        if c1 == 0:
            break
        n += c1 / (c2 - x_m2)
    return n


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_7_cython(cdouble w, double[::1] cs):
    cdef:
        double x_sqr = (2 * M_PI / creal(w)) ** 2
        double n = (
            cs[0]
            + cs[1] / (x_sqr - 0.028)
            + cs[2] / (x_sqr - 0.028) ** 2
            + cs[3] * x_sqr
            + cs[4] * x_sqr ** 2
            + cs[5] * x_sqr ** 3
        )
    return n


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_8_cython(cdouble w, double[::1] cs):
    cdef:
        double x_sqr = (2 * M_PI / creal(w)) ** 2
        double a = cs[0] + cs[1] * x_sqr / (x_sqr - cs[2]) + cs[3] * x_sqr
        double n_sqr = (1 + 2 * a) / (1 - a)
    return sqrt(n_sqr * (n_sqr > 0))


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_9_cython(cdouble w, double[::1] cs):
    cdef:
        double x = 2 * M_PI / creal(w)
        double n_sqr = (
                cs[0]
                + cs[1] / (x ** 2 - cs[2])
                + cs[3] * (x - cs[4]) / ((x - cs[4]) ** 2 + cs[5])
            )
    return sqrt(n_sqr * (n_sqr > 0))

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_21_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        cdouble _w = w * UNIT
        double eb = cs[0]
        double f0 = cs[1]
        double g0 = cs[2]
        double wp = cs[3]
        cdouble eps = eb - f0 * wp ** 2 / (_w ** 2 + C_I * _w * g0)
        double fj, gj, wj,

    for i in range(1, 7):
        fj = cs[3 * i + 1]
        gj = cs[3 * i + 2]
        wj = cs[3 * i + 3]
        if fj ==0:
            break
        eps -= fj * wp ** 2 / (_w ** 2 - wj ** 2 + C_I * _w * gj)
    return eps


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def formula_22_cython(cdouble w, double[::1] cs):
    cdef:
        int i
        cdouble _w = w * UNIT
        double eb = cs[0]
        double f0 = cs[1]
        double g0 = cs[2]
        double wp = cs[3]
        cdouble eps = eb - f0 * wp ** 2 / (_w ** 2 + 1j * _w * g0)
        cdouble aj
        double fj, gj, wj, sj, sj_inv

    for i in range(1, 6):
        fj = cs[4 * i]
        gj = cs[4 * i + 1]
        wj = cs[4 * i + 2]
        sj = cs[4 * i + 3]
        aj = csqrt(_w * (_w + 1j * gj))
        sj_inv = RT2_INV / sj if sj != 0 else 0
        if fj ==0:
            break
        eps += (
            C * fj * wp ** 2 / aj * sj_inv
            * (
                wofz((aj - wj) * sj_inv)
                + wofz((aj + wj) * sj_inv)
            )
        )
    return eps



import os

import numpy as np
from Cython.Distutils import build_ext
from setuptools import Extension, find_packages, setup

ext_modules = []
e = Extension(
    "riip.formulas_cython",
    sources=[os.path.join("src", "utils", "formulas_cython.pyx")],
    depends=[],
    include_dirs=[np.get_include(), "."],
    language="c++",
)
e.cython_directives = {"language_level": "3"}
ext_modules.append(e)

setup(
    name="riip",
    version="0.6.1",
    url="https://github.com/mnishida/RII_Pandas",
    license="MIT",
    author="Munehiro Nishida",
    author_email="mnishida@hiroshima-u.ac.jp",
    description="Python 3 + Pandas wrapper for the refractiveindex.info database",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    zip_safe=False,
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={"riip": ["py.typed", "data/riid.patch", "data/my_database"]},
    setup_requires=["Cython", "numpy", "scipy"],
    install_requires=[line.strip() for line in open("requirements.txt").readlines()],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Scientific/Engineering",
    ],
    ext_modules=ext_modules,
    cmdclass={"build_ext": build_ext},
)
